@@ -2,8 +2,9 @@
 use strict;
 use warnings;
 package YAML::PP;
+use B;
 
-our $VERSION = '0.005_001'; # TRIAL VERSION
+our $VERSION = '0.005_002'; # TRIAL VERSION
 
 use base 'Exporter';
 our @EXPORT_OK = qw/ Load LoadFile Dump DumpFile /;
@@ -13,6 +14,7 @@ sub new {
 
     my $bool = delete $args{boolean} // 'perl';
     my $schemas = delete $args{schema} || ['Core'];
+    my $cyclic_refs = delete $args{cyclic_refs} || 'allow';
 
     my $schema = YAML::PP::Schema->new(
         boolean => $bool,
@@ -22,11 +24,13 @@ sub new {
     my $self = bless {
         boolean => $bool,
         schema => $schema,
+        cyclic_refs => $cyclic_refs,
     }, $class;
     return $self;
 }
 
 sub boolean { return $_[0]->{boolean} }
+sub cyclic_refs { return $_[0]->{cyclic_refs} }
 
 sub loader {
     if (@_ > 1) {
@@ -66,8 +70,8 @@ sub load_string {
     unless ($loader) {
         require YAML::PP::Loader;
         $loader = YAML::PP::Loader->new(
-            boolean => $self->boolean,
             schema => $self->schema,
+            cyclic_refs => $self->cyclic_refs,
         );
         $self->loader($loader);
     }
@@ -80,7 +84,6 @@ sub load_file {
     unless ($loader) {
         require YAML::PP::Loader;
         $loader = YAML::PP::Loader->new(
-            boolean => $self->boolean,
             schema => $self->schema,
         );
         $self->loader($loader);
@@ -93,7 +96,9 @@ sub dump_string {
     my $dumper = $self->dumper;
     unless ($dumper) {
         require YAML::PP::Dumper;
-        $dumper = YAML::PP::Dumper->new;
+        $dumper = YAML::PP::Dumper->new(
+            schema => $self->schema,
+        );
         $self->dumper($dumper);
     }
     return $dumper->dump_string(@data);
@@ -104,7 +109,9 @@ sub dump_file {
     my $dumper = $self->dumper;
     unless ($dumper) {
         require YAML::PP::Dumper;
-        $dumper = YAML::PP::Dumper->new;
+        $dumper = YAML::PP::Dumper->new(
+            schema => $self->schema,
+        );
         $self->dumper($dumper);
     }
     return $dumper->dump_file($file, @data);
@@ -161,7 +168,8 @@ sub new {
     }
 
     my $self = bless {
-        resolvers => $args{resolvers} || {},
+        resolvers => {},
+        representers => {},
         true => $true,
         false => $false,
     }, $class;
@@ -169,6 +177,7 @@ sub new {
 }
 
 sub resolvers { return $_[0]->{resolvers} }
+sub representers { return $_[0]->{representers} }
 sub true { return $_[0]->{true} }
 sub false { return $_[0]->{false} }
 
@@ -182,34 +191,58 @@ sub load_subschemas {
     }
 }
 
-sub add_matcher {
-    my ($self, $type, $match, $value) = @_;
+sub add_resolver {
+    my ($self, %args) = @_;
+    my $tag = $args{tag};
+    my $rule = $args{match};
     my $resolvers = $self->resolvers;
-    my $res = $resolvers->{value} ||= {};
-    if ($type eq 'equals') {
-        unless (exists $res->{equals}->{ $match }) {
-            $res->{equals}->{ $match } = $value;
-        }
-        return;
+    my ($type, $match, $value) = @$rule;
+    my $implicit = $args{implicit} // 1;
+    my @resolvers;
+    if ($tag) {
+        my $res = $resolvers->{tag}->{ $tag } ||= {};
+        push @resolvers, $res;
     }
-    if ($type eq 'regex') {
-        push @{ $res->{regex} }, [ $match => $value ];
-        return;
+    if ($implicit) {
+        push @resolvers, $resolvers->{value} ||= {};
+    }
+    for my $res (@resolvers) {
+        if ($type eq 'equals') {
+            unless (exists $res->{equals}->{ $match }) {
+                $res->{equals}->{ $match } = $value;
+            }
+            next;
+        }
+        if ($type eq 'regex') {
+            push @{ $res->{regex} }, [ $match => $value ];
+        }
     }
 }
 
-sub add_tag_resolver {
-    my ($self, $tag, $type, $match, $value) = @_;
-    my $resolvers = $self->resolvers;
-    my $res = $resolvers->{tag}->{ $tag } ||= {};
-    if ($type eq 'equals') {
-        unless (exists $res->{equals}->{ $match }) {
-            $res->{equals}->{ $match } = $value;
-        }
+sub add_representer {
+    my ($self, %args) = @_;
+
+    my $representers = $self->representers;
+    if (my $flags = $args{flags}) {
+        my $rep = $representers->{flags} ||= [];
+        push @$rep, \%args;
         return;
     }
-    if ($type eq 'regex') {
-        push @{ $res->{regex} }, [ $match => $value ];
+    if (my $regex = $args{regex}) {
+        my $rep = $representers->{regex} ||= [];
+        push @$rep, \%args;
+        return;
+    }
+    if (defined(my $equals = $args{equals})) {
+        my $rep = $representers->{equals} ||= {};
+        $rep->{ $equals } = {
+            code => $args{code},
+        };
+        return;
+    }
+    if (my $undef = $args{undefined}) {
+        $representers->{undef} = $undef;
+        return;
     }
 }
 
@@ -300,14 +333,18 @@ use base 'YAML::PP::Schema';
 sub register {
     my ($self, %args) = @_;
     my $schema = $args{schema};
-    $schema->add_matcher( equals => '' => '' );
-    return {
-        tags => [
-#            '<tag:yaml.org,2002:str>' => {
-#                default => 1,
-#            },
-        ],
-    };
+
+    $schema->add_resolver(
+        match => [ equals => '' => '' ],
+    );
+
+    $schema->add_representer(
+        undefined => sub {
+            my ($rep, $value) = @_;
+            return { quoted => "" };
+        },
+    );
+    return;
 }
 
 package YAML::PP::Schema::JSON;
@@ -330,32 +367,80 @@ sub _to_float { unpack F => pack F => $_[0] }
 sub register {
     my ($self, %args) = @_;
     my $schema = $args{schema};
-    $schema->add_matcher( equals => null => undef );
-    $schema->add_matcher( equals => true => $schema->true );
-    $schema->add_matcher( equals => false => $schema->false );
-    $schema->add_matcher( regex => $RE_INT => \&_to_int );
-    $schema->add_matcher( regex => $RE_FLOAT => \&_to_float );
 
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:null' => equals => '' => undef
+    $schema->add_resolver(
+        tag => 'tag:yaml.org,2002:null',
+        match => [ equals => null => undef ],
     );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:null' => equals => null => undef
+    $schema->add_resolver(
+        tag => 'tag:yaml.org,2002:null',
+        match => [ equals => '' => undef ],
+        implicit => 0,
     );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:bool' => equals => true => $schema->true
+    $schema->add_resolver(
+        tag => 'tag:yaml.org,2002:bool',
+        match => [ equals => true => $schema->true ],
     );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:bool' => equals => false => $schema->false
+    $schema->add_resolver(
+        tag => 'tag:yaml.org,2002:bool',
+        match => [ equals => false => $schema->false ],
     );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:int' => regex => $RE_INT => \&_to_int
+    $schema->add_resolver(
+        tag => 'tag:yaml.org,2002:int',
+        match => [ regex => $RE_INT => \&_to_int ],
     );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:float' => regex => $RE_FLOAT => \&_to_float
+    $schema->add_resolver(
+        tag => 'tag:yaml.org,2002:float',
+        match => [ regex => $RE_FLOAT => \&_to_float ],
     );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:str' => regex => qr{^(.*)$} => sub { $_[0] }
+    $schema->add_resolver(
+        tag => 'tag:yaml.org,2002:str',
+        match => [ regex => qr{^(.*)$} => sub { $_[0] } ],
+        implicit => 0,
+    );
+
+    $schema->add_representer(
+        undefined => sub {
+            my ($rep, $value) = @_;
+            return { plain => "null" };
+        },
+    );
+    $schema->add_representer(
+        reftype => "*",
+        code => sub {
+            die "Dumping references not supported yet";
+        },
+    );
+
+    my $int_flags = B::SVp_IOK;
+    my $float_flags = B::SVp_NOK;
+    $schema->add_representer(
+        flags => $int_flags,
+        code => sub {
+            my ($rep, $value) = @_;
+            return { plain => "$value" };
+        },
+    );
+    $schema->add_representer(
+        flags => $float_flags,
+        code => sub {
+            my ($rep, $value) = @_;
+            return { plain => "$value" };
+        },
+    );
+    $schema->add_representer(
+        equals => $_,
+        code => sub {
+            my ($rep, $value) = @_;
+            return { quoted => "$value" };
+        },
+    ) for ("", qw/ true false null /);
+    $schema->add_representer(
+        regex => qr{$RE_INT|$RE_FLOAT},
+        code => sub {
+            my ($rep, $value) = @_;
+            return { quoted => "$value" };
+        },
     );
     return;
 }
@@ -381,70 +466,85 @@ sub _from_hex { hex $_[0] }
 sub register {
     my ($self, %args) = @_;
     my $schema = $args{schema};
-    $schema->add_matcher( equals => '' => undef );
-    $schema->add_matcher( equals => null => undef );
-    $schema->add_matcher( equals => NULL => undef );
-    $schema->add_matcher( equals => Null => undef );
-    $schema->add_matcher( equals => '~' => undef );
-    $schema->add_matcher( equals => true => $schema->true );
-    $schema->add_matcher( equals => TRUE => $schema->true );
-    $schema->add_matcher( equals => True => $schema->true );
-    $schema->add_matcher( equals => false => $schema->false );
-    $schema->add_matcher( equals => FALSE => $schema->false );
-    $schema->add_matcher( equals => False => $schema->false );
-    $schema->add_matcher( regex => $RE_INT_CORE => \&YAML::PP::Schema::JSON::_to_int );
-    $schema->add_matcher( regex => $RE_INT_OCTAL => \&_from_oct );
-    $schema->add_matcher( regex => $RE_INT_HEX => \&_from_hex );
-    $schema->add_matcher( regex => $RE_FLOAT_CORE => \&YAML::PP::Schema::JSON::_to_float );
 
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:null' => equals => '' => undef
+    $schema->add_resolver(
+        tag => 'tag:yaml.org,2002:null',
+        match => [ equals => $_ => undef ],
+    ) for (qw/ null NULL Null ~ /, '');
+    $schema->add_resolver(
+        tag => 'tag:yaml.org,2002:bool',
+        match => [ equals => $_ => $schema->true ],
+    ) for (qw/ true TRUE True /);
+    $schema->add_resolver(
+        tag => 'tag:yaml.org,2002:bool',
+        match => [ equals => $_ => $schema->false ],
+    ) for (qw/ false FALSE False /);
+    $schema->add_resolver(
+        tag => 'tag:yaml.org,2002:int',
+        match => [ regex => $RE_INT_CORE => \&YAML::PP::Schema::JSON::_to_int ],
     );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:null' => equals => null => undef
+    $schema->add_resolver(
+        tag => 'tag:yaml.org,2002:int',
+        match => [ regex => $RE_INT_OCTAL => \&_from_oct ],
     );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:null' => equals => '~' => undef
+    $schema->add_resolver(
+        tag => 'tag:yaml.org,2002:int',
+        match => [ regex => $RE_INT_HEX => \&_from_hex ],
     );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:null' => equals => Null => undef
+    $schema->add_resolver(
+        tag => 'tag:yaml.org,2002:float',
+        match => [ regex => $RE_FLOAT_CORE => \&YAML::PP::Schema::JSON::_to_float ],
     );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:null' => equals => NULL => undef
+    $schema->add_resolver(
+        tag => 'tag:yaml.org,2002:str',
+        match => [ regex => qr{^(.*)$} => sub { $_[0] } ],
+        implicit => 0,
     );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:bool' => equals => true => $schema->true
+
+    $schema->add_representer(
+        reftype => "*",
+        code => sub {
+            die "Dumping references not supported yet";
+        },
     );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:bool' => equals => True => $schema->true
+
+    my $int_flags = B::SVp_IOK;
+    my $float_flags = B::SVp_NOK;
+    $schema->add_representer(
+        flags => $int_flags,
+        code => sub {
+            my ($rep, $value) = @_;
+            return { plain => "$value" };
+        },
     );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:bool' => equals => TRUE => $schema->true
+    $schema->add_representer(
+        flags => $float_flags,
+        code => sub {
+            my ($rep, $value) = @_;
+            return { plain => "$value" };
+        },
     );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:bool' => equals => false => $schema->false
+    $schema->add_representer(
+        undefined => sub {
+            my ($rep, $value) = @_;
+            return { plain => "null" };
+        },
     );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:bool' => equals => False => $schema->false
-    );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:bool' => equals => FALSE => $schema->false
-    );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:int' => regex => $RE_INT_CORE => \&YAML::PP::Schema::JSON::_to_int
-    );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:int' => regex => $RE_INT_OCTAL => \&_from_oct
-    );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:int' => regex => $RE_INT_HEX => \&_from_hex
-    );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:float' => regex => $RE_FLOAT_CORE => \&YAML::PP::Schema::JSON::_to_float
-    );
-    $schema->add_tag_resolver(
-        'tag:yaml.org,2002:str' => regex => qr{^(.*)$} => sub { $_[0] }
+    $schema->add_representer(
+        equals => $_,
+        code => sub {
+            my ($rep, $value) = @_;
+            return { quoted => "$value" };
+        },
+    ) for ("", qw/ true TRUE True false FALSE False null NULL Null ~ /);
+    $schema->add_representer(
+        regex => qr{$RE_INT_CORE|$RE_FLOAT_CORE|$RE_INT_OCTAL|$RE_INT_HEX},
+        code => sub {
+            my ($rep, $value) = @_;
+            return { quoted => "$value" };
+        },
     );
+
     return;
 }
 
@@ -541,7 +641,11 @@ The Loader combines the constructor and parser.
 
 =item L<YAML::PP::Dumper>
 
-The Dumper will create Emitter events from the given data structure.
+The Dumper will delegate to the Representer
+
+=item L<YAML::PP::Representer>
+
+The Representer will create Emitter events from the given data structure.
 
 =item L<YAML::PP::Emitter>
 
@@ -577,6 +681,10 @@ might still be wrong in some cases.
 
 The error messages need to be improved.
 
+=item Unicode Surrogate Pairs
+
+Currently loaded as single characters without validating
+
 =item Possibly more
 
 =back
@@ -667,12 +775,23 @@ wish to be able to keep the comments.
 
 =head2 YAML::PP::Dumper, YAML::PP::Emitter
 
-This is also pretty simple so far. Any string containing something
-other than C<0-9a-zA-Z.-> will be dumped with double quotes.
+The Dumper should be able to dump strings correctly, adding quotes
+whenever a plain scalar would look like a special string, like C<true>,
+or when it contains or starts with characters that are not allowed.
+
+Most strings will be dumped as plain scalars without quotes. If they
+contain special characters or have a special meaning, they will be dumped
+with single quotes. If they contain control characters, including <"\n">,
+they will be dumped with double quotes.
 
 It will recognize JSON::PP::Boolean and boolean.pm objects and dump them
 correctly.
 
+TODO: Correctly recognize numbers which also have a string flag, like:
+
+    my $int = 23;
+    say "int: $int"; # $int will now also have a PV flag
+
 The layout is like libyaml output:
 
     key:
@@ -689,6 +808,69 @@ The layout is like libyaml output:
     - - b1
       - b2
 
+=head1 METHODS
+
+=over
+
+=item new
+
+    my $ypp = YAML::PP->new;
+    # load booleans via boolean.pm
+    my $ypp = YAML::PP->new( boolean => 'boolean' );
+    # load booleans via JSON::PP::true/false
+    my $ypp = YAML::PP->new( boolean => 'JSON::PP' );
+    
+    # use YAML 1.2 Failsafe Schema
+    my $ypp = YAML::PP->new( schema => ['Failsafe'] );
+    # use YAML 1.2 JSON Schema
+    my $ypp = YAML::PP->new( schema => ['JSON'] );
+    # use YAML 1.2 Core Schema
+    my $ypp = YAML::PP->new( schema => ['Core'] );
+    
+    # Die when detecting cyclic references
+    my $ypp = YAML::PP->new( cyclic_refs => 'fatal' );
+    # Other values:
+    # warn   - Just warn about them and replace with undef
+    # ignore - replace with undef
+    # allow  - Default
+    
+    my $ypp = YAML::PP->new(
+        boolean => 'JSON::PP',
+        schema => ['JSON'],
+        cyclic_refs => 'fatal',
+    );
+
+=item load_string
+
+    my $doc = $ypp->load_string("foo: bar");
+    my @docs = $ypp->load_string("foo: bar\n---\n- a");
+
+Input should be utf-8 decoded.
+
+=item load_file
+
+    my $doc = $ypp->load_file("file.yaml");
+    my @docs = $ypp->load_file("file.yaml");
+
+UTF-8 decoding will be done automatically
+
+=item dump_string
+
+    my $yaml = $ypp->dump_string($doc);
+    my $yaml = $ypp->dump_string($doc1, $doc2);
+    my $yaml = $ypp->dump_string(@docs);
+
+Output will be UTF-8 decoded
+
+=item dump_file
+
+    $ypp->dump_file("file.yaml", $doc);
+    $ypp->dump_file("file.yaml", $doc1, $doc2);
+    $ypp->dump_file("file.yaml", @docs);
+
+File will be written UTF-8 encoded
+
+=back
 
 =head1 NUMBERS
 
