@@ -3,7 +3,7 @@ use strict;
 use warnings;
 package YAML::PP::Loader;
 
-our $VERSION = '0.005_001'; # TRIAL VERSION
+our $VERSION = '0.005_002'; # TRIAL VERSION
 
 use YAML::PP::Parser;
 use YAML::PP::Constructor;
@@ -12,14 +12,15 @@ use YAML::PP::Reader;
 sub new {
     my ($class, %args) = @_;
 
-    my $bool = delete $args{boolean} // 'perl';
+    my $cyclic_refs = delete $args{cyclic_refs} || 'allow';
     my $schema = delete $args{schema} // YAML::PP->default_schema(
-        boolean => $bool,
+        boolean => 'perl',
     );
 
     my $parser = delete $args{parser} || YAML::PP::Parser->new;
     my $constructor = delete $args{constructor} || YAML::PP::Constructor->new(
         schema => $schema,
+        cyclic_refs => $cyclic_refs,
     );
     if (keys %args) {
         die "Unexpected arguments: " . join ', ', sort keys %args;
